/* written by John Doug Reynolds, March 2001 */

#include "rgsfilter.h"
#include <ExposureMapper.h>
#include <EventAttributes.h>
#include <DataFilter.h>
#include <OGIPgti.h>
#include <cstdio>

namespace {
   using namespace rgsPixelsLib;

   class ExposureCombiner {
   public:
      ~ExposureCombiner();

      ExposureCombiner( const string &filename, const DataSet *eventset,
			const GridSpan &beta, const GridSpan &xdsp );

      void load( const PixelMap<float>&,
		 const char *extname, unsigned ccd, unsigned node );

   private:
      struct DataSet* _dataset;
      struct Array* _array;
      struct ArrayData* _arrayData;
      struct PixelMap<float> _map;
      struct Span _column, _nodeid;
   };
}

static void addExpMaps(struct DataSet*,bool withcombmap,const string &combmap);

void rgsfilter ()
{
   string dsName = stringParameter("set");
   string fsName = stringParameter("filteredset");

   int rejflags;
   {
      using namespace EventAttributes;

      rejflags = CLEAR_ALL_FLAGS;
      vector<string>::const_iterator itr;
      vector<string> evattrList = stringListParameter("rejflags");

      for ( itr = evattrList.begin(); itr != evattrList.end(); ++itr )
	 if ( *itr == "BAD_SHAPE" )
	    setEventAttribute( rejflags, BAD_SHAPE );
	 else if ( *itr == "ON_BADPIX" )
	    setEventAttribute( rejflags, ON_BADPIX );
	 else if ( *itr == "NEXT_TO_BADPIX" )
	    setEventAttribute( rejflags, NEXT_TO_BADPIX );
	 else if ( *itr == "NEXT_TO_CCD_BORDER" )
	    setEventAttribute( rejflags, NEXT_TO_CCD_BORDER );
	 else
	    Sas::error("badParameter");
   }

   struct DataSet *dataset = dataSetServer->clone(dsName,fsName,Dal::Modify);

   // merge the external GTI table

   if ( booleanParameter("withauxgti") ) {
      string gtitable = stringParameter("gtitable");

      struct Table *table = dataset->table("STDGTI");
      struct InOGIPgtiTable stdTable(table), usrTable(gtitable);
      struct GtiList stdGTI, usrGTI;
      stdTable >> stdGTI;
      usrTable >> usrGTI;

      struct OutOGIPgtiTable merged(table);
      merged << (stdGTI &= usrGTI);
   }

   // filter the EVENTS table

   struct Table *etab = dataset->table("EVENTS");
   
   char expression[64];
   sprintf( expression, "GTI([STDGTI],TIME) && (FLAG & %d == 0)", rejflags );
   DataFilter(expression).doFilter(*etab);

   if ( !etab->hasAttribute("REJFLAGS") )
      *etab->addAttribute("REJFLAGS","flags already filtered","XMMEA") = 0;

   struct Attribute &rejflagsAttr = *etab->attribute("REJFLAGS");
   rejflagsAttr = rejflagsAttr.asInt() | rejflags;

   // filter the EXPOSURE and REJPIX tables

   for ( unsigned ccd = 1; ccd <= 9; ++ccd ) {
      char xtabName[9], rtab0Name[9], rtab1Name[9];

      sprintf( xtabName, "EXPOSU0%d", ccd );
      if ( !dataset->hasBlock(xtabName) ) continue;

      sprintf( rtab0Name, "REJPIX0%d", ccd );
      sprintf( rtab1Name, "REJPIX1%d", ccd );

      struct Table *xtab = dataset->table(xtabName);
      struct Table *rtab0 = dataset->table(rtab0Name);
      struct Table *rtab1 = dataset->table(rtab1Name);

      DataFilter("GTI([STDGTI],TIME)").doFilter(*xtab);

      unsigned r0Count = 0, r1Count = 0;
      {
	 const DalVecJ xFrame( xtab->column("FRAME") );
	 const DalVecJ r0Frame( rtab0->column("FRAME") );
	 const DalVecJ r1Frame( rtab1->column("FRAME") );

	 unsigned xItr = 0, r0Itr = 0, r1Itr = 0;
	 while ( xItr < xtab->rows() ) {
	    int frame = xFrame[xItr++];

	    while ( r0Itr < rtab0->rows() && r0Frame[r0Itr] < frame ) ++r0Itr;
	    while ( r0Itr < rtab0->rows() && r0Frame[r0Itr] == frame ) {
	       if ( r0Itr != r0Count ) rtab0->copyRows(r0Itr,r0Count);
	       ++r0Itr, ++r0Count;
	    }

	    while ( r1Itr < rtab1->rows() && r1Frame[r1Itr] < frame ) ++r1Itr;
	    while ( r1Itr < rtab1->rows() && r1Frame[r1Itr] == frame ) {
	       if ( r1Itr != r1Count ) rtab1->copyRows(r1Itr,r1Count);
	       ++r1Itr, ++r1Count;
	    }
	 }
      }
      rtab0->deleteRows( r0Count, rtab0->rows() - r0Count );
      rtab1->deleteRows( r1Count, rtab1->rows() - r1Count );
   }

   // construct the filtered exposure maps

   if ( booleanParameter("withexpmaps") )
      addExpMaps( dataset,
		  booleanParameter("withcombmap"),
		  stringParameter("expimageset")
		  );

   dataSetServer->close( dataset );
}

void addExpMaps( struct DataSet *dataset,
		 bool withcombmap, const string &combmap )
{
   const Table *etab	= dataset->table("EVENTS");
   struct GridSpan beta = gridspan( etab->column("BETA_CHANNEL") );
   struct GridSpan xdsp = gridspan( etab->column("XDSP_CHANNEL") );
   float aspfact1	= etab->attribute("ASPFACT1")->asReal();
   int   rejflags	= etab->attribute("REJFLAGS")->asInt();

   string instrument	= dataset->attribute("INSTRUME")->asString();
   string datamode	= dataset->attribute("DATAMODE")->asString();
   int ccdocb		= etab->attribute("CCDOCB")->asInt();

   calServer.setInstrument( Xmm::toInstrument(instrument) );
   calServer.state()->mode()->set( Xmm::toMode(datamode) );
   calServer.state()->ocb()->set(ccdocb);

   struct ExposureCombiner *combiner = 0;
   if ( withcombmap )
      combiner = new ExposureCombiner(combmap,dataset,beta,xdsp);

   for ( unsigned ccd = 1; ccd <= 9; ++ccd ) {
      calServer.state()->ccd()->set(ccd);
      struct GeometryDataServer *calGDS = calServer.getAtom(calGDS);

      char xtabName[9];
      sprintf( xtabName, "EXPOSU0%d", ccd );
      if ( !dataset->hasBlock(xtabName) ) continue;

      struct Table *xtab = dataset->table(xtabName);
      const DalVecE timedel( xtab->column("TIMEDEL") );
      struct ExposureMapper mapper( xtab, aspfact1, beta, xdsp, calGDS );

      for ( unsigned node = 0; node < 2; ++node ) {
	 char extname[9], btabName[9], rtabName[9];
	 sprintf( extname, "EXPMAP%d%d", node, ccd );
	 sprintf( btabName, "BADPIX%d%d", node, ccd );
	 sprintf( rtabName, "REJPIX%d%d", node, ccd );

	 errstr << "adding extension " << extname
		<< Sas::message(Msg::App,Msg::Sparse);

	 const Table *btab = dataset->table(btabName);
	 const Table *rtab = dataset->table(rtabName);
	 mapper.load(btab,rtab,rejflags);

	 longVec size(2);
	 size[0] = mapper.bounds.row.length();
	 size[1] = mapper.bounds.col.length();

	 if ( dataset->hasBlock(extname) ) dataset->deleteBlock(extname);
	 struct Array *array = dataset->addArray( extname, Dal::Real32, size,
						  "s", "exposure(beta,xdsp)" );

	 struct Pixel origin( mapper.bounds.col.min, mapper.bounds.row.min );

#define SET(name,value,label) *array->addAttribute(name,label) = value

	 SET( "CTYPE1", "", "[dispersion] channel axis type" );
	 SET( "CUNIT1", "rad", "channel axis units" );
	 SET( "CRPIX1", 1, "reference channel coordinate" );
	 SET( "CRVAL1", beta[origin.col], "reference physical coordinate" );
	 SET( "CDELT1", beta.stride, "uniform channel width" );

	 SET( "CTYPE2", "", "[cross-dispersion] channel axis type" );
	 SET( "CUNIT2", "rad", "channel axis units" );
	 SET( "CRPIX2", 1, "reference channel coordinate" );
	 SET( "CRVAL2", xdsp[origin.row], "reference physical coordinate" );
	 SET( "CDELT2", xdsp.stride, "uniform channel width" );

	 SET( "CCDNODE", (int)node, "readout node identifier" );
	 SET( "CCDID", (int)ccd, "CCD identifier" );
	 array->addAttribute( etab->attribute("CCDOCB") );
	 array->addAttribute( etab->attribute("SOURCEID") );

	 SET( "T_BETA_X", mapper.tBetaX, "transform: chipX per beta bin" );
	 SET( "T_XDSP_X", mapper.tXdspX, "transform: chipX per xdsp bin" );
	 SET( "T_ORIG_X", mapper.tOrigX, "transform: chipX at origin" );
	 SET( "T_BETA_Y", mapper.tBetaY, "transform: chipY per beta bin" );
	 SET( "T_XDSP_Y", mapper.tXdspY, "transform: chipY per xdsp bin" );
	 SET( "T_ORIG_Y", mapper.tOrigY, "transform: chipY at origin" );

	 struct ArrayData *arrayData = array->data();
	 float *buffer = arrayData->real32Data();
	 size_t bufferSize = size[0] * size[1];
	 memset( buffer, 0, bufferSize * sizeof(float) );

	 struct PixelMap<float> xmap( mapper.bounds, 0, buffer );
	 const DalVecE fracexp( xtab->column(node?"FRACEXP1":"FRACEXP0") );
	 for ( unsigned i = 0; i < xtab->rows(); ++i )
	    mapper.expose( xmap, i, timedel[i] * fracexp[i] );

	 float datamin, datamax;
	 for ( datamin = datamax = *buffer; bufferSize--; ++buffer ) {
	    minimize( datamin, *buffer );
	    maximize( datamax, *buffer );
	 }

	 SET( "DATAMIN", datamin, "minimum data value present" );
	 SET( "DATAMAX", datamax, "maximum data value present" );

#undef SET

	 if ( combiner ) combiner->load( xmap, extname, ccd, node );
	 delete arrayData;
      }

      calServer.releaseAtom(calGDS);
   }

   delete combiner;
}

/* --- ExposureCombiner methods -------------------------------------------- */

ExposureCombiner::
ExposureCombiner( const string &filename, const DataSet *eventset,
		  const GridSpan &beta, const GridSpan &xdsp )
{
   _dataset = dataSetServer->open( filename, Dal::Create );
   
   _dataset->addAttribute( eventset->attribute("TELESCOP") );
   _dataset->addAttribute( eventset->attribute("INSTRUME") );
   _dataset->addAttribute( eventset->attribute("OBS_ID") );
   _dataset->addAttribute( eventset->attribute("EXP_ID") );
   _dataset->addAttribute( eventset->attribute("DATE-OBS") );
   _dataset->addAttribute( eventset->attribute("DATE-END") );

   longVec size(2);
   size[0] = xdsp.length();
   size[1] = beta.length();

   _array = _dataset->addArray( "PRIMARY", Dal::Real32, size,
				"s", "exposure(beta,xdsp)" );

#define SET(name,value,label) *_array->addAttribute(name,label) = value

   SET( "CTYPE1", "", "[dispersion] channel axis type" );
   SET( "CUNIT1", "rad", "channel axis units" );
   SET( "CRPIX1", 1, "reference channel coordinate" );
   SET( "CRVAL1", beta[1], "reference physical coordinate" );
   SET( "CDELT1", beta.stride, "uniform channel width" );

   SET( "CTYPE2", "", "[cross-dispersion] channel axis type" );
   SET( "CUNIT2", "rad", "channel axis units" );
   SET( "CRPIX2", 1, "reference channel coordinate" );
   SET( "CRVAL2", xdsp[1], "reference physical coordinate" );
   SET( "CDELT2", xdsp.stride, "uniform channel width" );

#undef SET

   const Table *events = eventset->table("EVENTS");
   _array->addAttribute( events->attribute("CCDOCB") );
   _array->addAttribute( events->attribute("SOURCEID") );

   _arrayData = _array->data();
   float *buffer = _arrayData->real32Data();
   memset( buffer, 0, size[0] * size[1] * sizeof(float) );
   _map.resize( Bounds(beta,xdsp), 0, buffer );
}

void ExposureCombiner::load( const PixelMap<float> &xmap,
			     const char *name, unsigned ccd, unsigned node )
{
   struct Bounds bounds = xmap.bounds() & _map.bounds();

   if ( !bounds ) {
      errstr << name << " does not overlap the common channel space"
	     << Sas::warning("lostExposure");
      return;
   }

   if ( xmap.bounds().col.min < bounds.col.min )
      errstr << name << " extends " << bounds.col.min - xmap.bounds().col.min
	     << " columns left of the common channel space"
	     << Sas::warning("lostExposure");

   if ( xmap.bounds().col.max > bounds.col.max )
      errstr << name << " extends " << xmap.bounds().col.max - bounds.col.max
	     << " columns right of the common channel space"
	     << Sas::warning("lostExposure");

   if ( xmap.bounds().row.min < bounds.row.min )
      errstr << name << " extends " << bounds.row.min - xmap.bounds().row.min
	     << " rows below the common channel space"
	     << Sas::warning("lostExposure");

   if ( xmap.bounds().row.max > bounds.row.max )
      errstr << name << " extends " << xmap.bounds().row.max - bounds.row.max
	     << " rows above the common channel space"
	     << Sas::warning("lostExposure");

   for ( int row = bounds.row.min; row <= bounds.row.max; ++row ) {
      using Compass::East;
      long src = xmap.id( bounds.col.min - 1, row );
      long dst = _map.id( bounds.col.min - 1, row );
      for ( size_t count = bounds.col.length(); count--; )
	 _map[ _map.incr(dst,East) ] += xmap[ xmap.incr(src,East) ];
   }

   _nodeid.extend( 2 * ccd + node - 2 );
   _column.extend( xmap.bounds().col.min );
   _column.extend( xmap.bounds().col.max );
}

ExposureCombiner::~ExposureCombiner()
{
#define SET(name,value,label) *_array->addAttribute(name,label) = value

   float slope = -((float)_nodeid.length()) / _column.length();
   float icept = 1 + _nodeid.max - slope * (_column.min - .5);

   SET( "T_BETA_N", slope, "transform: nodes per beta bin" );
   SET( "T_ORIG_N", icept, "transform: node offset at origin" );

   float datamin, datamax, *buffer = &_map[0];
   size_t bufferSize = _map.bounds().col.length() * _map.bounds().row.length();
   for ( datamin = datamax = *buffer; bufferSize--; ++buffer ) {
      minimize( datamin, *buffer );
      maximize( datamax, *buffer );
   }

   SET( "DATAMIN", datamin, "minimum data value present" );
   SET( "DATAMAX", datamax, "maximum data value present" );

   delete _arrayData;
   dataSetServer->close( _dataset );

#undef SET
}
